@@ -34,7 +34,7 @@ if __name__ == "__main__" :
     multivac.init(_session_maker)
     
     base.BaseTest._session_maker = _session_maker
-    multivac.metadata.create_all()
+    multivac.metadata.create_all(engine)
     
     te = test_element.TestElement()
     te.setup()
