@@ -3,12 +3,16 @@
 
 _multiprocess_can_split_ = True
 
+import warnings
+
 from nose.tools import raises
-from sqlalchemy.exc import IntegrityError
+from nose.exc import SkipTest
+from sqlalchemy.exc import IntegrityError, UnboundExecutionError, InvalidRequestError
 from test_project import TestProject
 
+import multivac
+
 from multivac import Element
-from multivac import Project
 
 class TestElementBase(TestProject):
     pass
@@ -24,7 +28,7 @@ class TestCreate(TestElementBase):
     def test_id(self, **kwargs):
         @raises(AttributeError)
         def check_id(field):
-            p = Element(**dict({field: self.randint()}))
+            Element(**dict({field: self.randint()}))
             # The previous statement should have failed
             assert False
         
@@ -166,6 +170,34 @@ class TestUpdate(TestElementBase):
         assert e is r
         assert e.status == r.status
 
+class TestHistory(TestElementBase):
+    def test_insert(self, **kwargs):
+        e = TestCreate().test_required()
+        self.session.add(e)
+        self.session.flush()
+        assert len(e.status_history) == 1
+        assert e.status_history[0].status == e.status
+    
+    def test_update(self, **kwargs):
+        e = TestCreate().test_required(status='2_STATUS')
+        self.session.add(e)
+        self.session.flush()
+        e.status = '1_STATUS'
+        self.session.flush()
+        assert len(e.status_history) == 2
+        assert e.status_history[0].status == '2_STATUS'
+        assert e.status_history[1].status == '1_STATUS'
+    
+    @raises(InvalidRequestError)
+    def test_delete(self, **kwargs):
+        e = TestCreate().test_required()
+        self.session.add(e)
+        self.session.flush()
+        s = e.status_history[0]
+        self.session.delete(e)
+        self.session.flush()
+        self.session.refresh(s)
+
 class TestDelete(TestElementBase):
     def test_delete(self):
         e = TestCreate().test_required()
@@ -259,7 +291,7 @@ class TestTag(TestElementBase):
         assert 't' not in e.tag
 
 class TestElement(TestElementBase):
-    """
+    u"""
        Tree representation of all elements classified by project
        
                          [p1]                                            │                   [p2]                                            │                   [p3]                                    │                   [p4]                                    │                   [p5]                                  
@@ -411,8 +443,8 @@ class TestQuery(TestElement):
         assert self.element['e2322'] in r
         assert self.element['e2632'] in r
     
-    def test_tag_true(self):
-        r = Element.query_tags(self.session, tags={'t5': True}).all()
+    def test_tag_key(self):
+        r = self.session.query(Element).filter(Element.tags.any(name='t5')).all() # @UndefinedVariable
         assert len(r) == 13
         assert self.element['e1533'] in r
         assert self.element['e1745'] in r
@@ -428,15 +460,15 @@ class TestQuery(TestElement):
         assert self.element['e5423'] in r
         assert self.element['e5533'] in r
     
-    def test_tag_none(self):
-        r = Element.query_tags(self.session, tags={'t5': None}).all()
+    def test_tag_key_none(self):
+        r = self.session.query(Element).filter(Element.tags.any(name='t5', value=None)).all() # @UndefinedVariable
         assert len(r) == 3
         assert self.element['e1745'] in r
         assert self.element['e2745'] in r
         assert self.element['e3745'] in r
         
-    def test_tag_value(self):
-        r = Element.query_tags(self.session, tags={'t5': 'e3533t5'}).filter_by(name='e3533').all()
+    def test_tag_key_value(self):
+        r = self.session.query(Element).filter(Element.tags.any(name='t5', value='e3533t5')).all() # @UndefinedVariable
         assert len(r) == 1
         assert self.element['e3533'] in r
 
@@ -589,8 +621,6 @@ class TestStatusCascade(TestElement):
         #     e1843.parent = None
         # Postcondition:
         #     e1423 (3F) -> e1745 (5)
-        #import pdb
-        #pdb.set_trace()
         assert self.element['e1423'].status == '3_STATUS'
         assert self.element['e1745'].status == '5_STATUS'
         assert self.element['e1843'].status == '3_STATUS'
@@ -635,3 +665,68 @@ class TestStatusCascade(TestElement):
         assert self.element['e1322'].status == '2_STATUS'
         assert self.element['e1423'].status == '3_STATUS'
         assert self.element['e1202'].forced_status == False
+
+class TestCTE(TestElement):
+    def test_ancestor_no_parent(self):
+        assert self.element['e2202'].ancestors == []
+    
+    @raises(UnboundExecutionError)
+    def test_disable_bind(self):
+        self.session.flush()
+        bind = self.session.bind
+        try:
+            # Disable statement execution
+            self.session.bind = None
+            multivac.metadata.bind = None
+            # Raise an error
+            self.session.query(Element).first()
+        finally:
+            # Reenable statement execution
+            self.session.bind = bind
+            multivac.metadata.bind = bind
+    
+    def test_ancestor_in_session(self):
+        # Force flush before its disallowed.
+        self.session.flush()
+        # Save engine
+        bind = self.session.bind
+        try:
+            # Disable statement execution
+            self.session.bind = None
+            multivac.metadata.bind = None
+            # Retrieve and verify the ancestors
+            a = self.element['e2533'].ancestors        
+            assert a[0] == self.element['e2322']
+            assert a[1] == self.element['e2202']
+        finally:
+            # Reenable statement execution
+            self.session.bind = bind
+            multivac.metadata.bind = bind
+    
+    def test_ancestor_without_cte(self):
+        if self.session.bind.name == 'postgresql':
+            raise SkipTest('This database uses CTEs.')
+        
+        with warnings.catch_warnings():
+            warnings.filterwarnings("ignore", category=multivac.CTENotSupported)
+            self.session.flush()
+            self.session.expunge_all()
+            # After expunging all objects from the session, we can no longer use them
+            e = self.session.query(Element).filter_by(name='e2533').one()
+            a = list(e.ancestors)
+            assert a[0].name == 'e2322'
+            assert a[1].name == 'e2202'
+    
+    def test_ancestor_with_cte(self):
+        if self.session.bind.name != 'postgresql':
+            raise SkipTest('CTEs are only supported on PostgreSQL.')  
+        
+        with warnings.catch_warnings():
+            warnings.filterwarnings("ignore", category=multivac.CTENotSupported)
+            self.session.flush()
+            self.session.expunge_all()
+            # After expunging all objects from the session, we can no longer use them
+            e = self.session.query(Element).filter_by(name='e2533').one()
+            a = list(e.ancestors)
+            assert a[0].name == 'e2322'
+            assert a[1].name == 'e2202'
